#include <cuda.h>
#include <cuda_runtime_api.h>
#include <device_launch_parameters.h>
#include <device_atomic_functions.h>
#include <cstdio>
#include "defs.h"
#include <mpi.h>

#define BLOCK_SIZE 32LL
#define THREADS_PER_BLOCK 512
#define SHIFTBID(dx, dy) (bucket_id + dy + dx * dev_n_buckets_y)
#define GET_CELL_ELEMENT(bucket_id) ((unsigned int *)((char*)buckets + bucket_id * pitch))

#define CUDA_CALL(x) do { if((x) != cudaSuccess) { \
    printf("CUDA-Error at %s:%d: '%s'\n",__FILE__,__LINE__, cudaGetErrorString(x)); \
    exit(x);}} while(0)

extern "C" void print_timing(const char* full_string, const char* short_string, const double value);
particle_t* dev_particles;
pair_t* dev_pairs;
unsigned int *dev_pair_count;
real_t* dev_subdomain;
int* dev_interactions;

#ifdef BUCKET
__constant__ int_t dev_n_buckets_x;
__constant__ int_t dev_n_buckets_y;
#endif

// Timing 
double
    t_find_pairs = 0.0,
    t_transfer = 0.0,
    t_sort = 0.0,
    t_setup = 0.0,
    t_find_neighbors = 0.0,
    t_physics = 0.0,
    t_update_pairs = 0.0,
    t_alloc = 0.0;

int_t n_pairs = 0;
int_t n_pair_cap = CAP_INCREMENT;
size_t dev_particles_max = (1+L/DELTA) * (1+T/DELTA);
//pair_t* dev_pairs;

#if defined(BUCKET)
    int_t n_buckets = 0;
#endif

#if !defined(BUCKET)

__global__ void naive(const int_t n_total, particle_t* particles, pair_t *all_pairs,
	unsigned int* pair_count, int* interactions) {
    const int_t i = blockDim.x * blockIdx.x + threadIdx.x;
    const int_t j = blockDim.y * blockIdx.y + threadIdx.y;

    if (i == j || i >= n_total - 1 || j <= i || j >= n_total)
        return;

    const real_t difx = particles[i].x[0] - particles[j].x[0];
    const real_t dify = particles[i].x[1] - particles[j].x[1];
    if (difx*difx + dify * dify < POW2(scale_k*H)) {
        // Is pair
        unsigned int kk = atomicAdd(pair_count, 1);
        interactions[i] += 1;
        interactions[j] += 1;
        all_pairs[kk].i = i;
        all_pairs[kk].j = j;
        all_pairs[kk].r = sqrt(difx*difx + dify * dify);
        all_pairs[kk].q = sqrt(difx*difx + dify * dify) / H;
        all_pairs[kk].w = 0.0f;
        all_pairs[kk].dwdx[0] = all_pairs[kk].dwdx[1] = 0.0f;
    }
}

#else

unsigned int** buckets;
unsigned int* bucket_particle_count;
size_t pitch;

__global__ void fill_buckets(const int_t n_total, particle_t* particles, unsigned int ** buckets, size_t pitch, unsigned int* bucket_particle_count, int_t n_buckets) {
    const int_t i = blockDim.x * blockIdx.x + threadIdx.x;
    if (i >= n_total)
        return;
    int_t bucket_id = particles[i].bucket_id;
    if (bucket_id < 0 || bucket_id >= n_buckets) {
        return;
    }

    unsigned int new_index = atomicAdd(&bucket_particle_count[bucket_id], 1);
    if (new_index < INITIAL_MAX_BUCKET) {
        unsigned int* bucket = GET_CELL_ELEMENT(bucket_id);
        bucket[new_index] = i;
    }
}

__device__ inline void compare(int_t bucket_id, int_t n_buckets, unsigned int** buckets, particle_t* particles, particle_t particle, int* interactions, pair_t * pairs, unsigned int* pair_count, size_t pitch, unsigned int* bucket_particle_count) {
    if (bucket_id < 0 || bucket_id >= n_buckets)
        return;
    const unsigned int total_p = min(bucket_particle_count[bucket_id], INITIAL_MAX_BUCKET-1);
    unsigned int* bucket = GET_CELL_ELEMENT(bucket_id);
    for (unsigned int i = 0; i < total_p; i++) {
        particle_t other_particle = particles[bucket[i]];
        if (other_particle.local_idx < particle.local_idx) {
            const real_t difx = particle.x[0] - other_particle.x[0];
            const real_t dify = particle.x[1] - other_particle.x[1];
            if (difx*difx + dify * dify < POW2(scale_k*H)) {
                interactions[particle.local_idx]++;
                interactions[other_particle.local_idx]++;

                pair_t loc_pair;
                loc_pair.i = particle.local_idx;
                loc_pair.j = other_particle.local_idx;
                loc_pair.r = sqrt(difx*difx + dify * dify);
                loc_pair.q = sqrt(difx*difx + dify * dify) / H;
                loc_pair.w = 0.0f;
                loc_pair.dwdx[0] = loc_pair.dwdx[1] = 0.0f;
                unsigned int pair_idx = atomicAdd(pair_count, 1);
                pairs[pair_idx] = loc_pair;
            }
        }
    }
}

__global__ void
__launch_bounds__(THREADS_PER_BLOCK)
 find_pairs(unsigned int** buckets, int_t n_buckets, particle_t* particles, int_t n_particles, int* interactions, pair_t * pairs, unsigned int* pair_count, size_t pitch, unsigned int* bucket_particle_count) {
    const int_t p_id = blockDim.x * blockIdx.x + threadIdx.x;
    if (p_id >= n_particles)
        return;

    particle_t particle = particles[p_id];
    const int bucket_id = particle.bucket_id;
    // Center / local
    compare(bucket_id, n_buckets, buckets, particles, particle, interactions, pairs, pair_count, pitch, bucket_particle_count);
    // North West
    compare(SHIFTBID(-1, 1), n_buckets, buckets, particles, particle, interactions, pairs, pair_count, pitch, bucket_particle_count);
    // North
    compare(SHIFTBID(0, 1), n_buckets, buckets, particles, particle, interactions, pairs, pair_count, pitch, bucket_particle_count);
    // North East
    compare(SHIFTBID(1, 1), n_buckets, buckets, particles, particle, interactions, pairs, pair_count, pitch, bucket_particle_count);
    // East
    compare(SHIFTBID(1, 0), n_buckets, buckets, particles, particle, interactions, pairs, pair_count, pitch, bucket_particle_count);
    // South East
    compare(SHIFTBID(1, -1), n_buckets, buckets, particles, particle, interactions, pairs, pair_count, pitch, bucket_particle_count);
    // South
    compare(SHIFTBID(0, -1), n_buckets, buckets, particles, particle, interactions, pairs, pair_count, pitch, bucket_particle_count);
    // South West
    compare(SHIFTBID(-1, -1), n_buckets, buckets, particles, particle, interactions, pairs, pair_count, pitch, bucket_particle_count);
    // West
    compare(SHIFTBID(-1, 0), n_buckets, buckets, particles, particle, interactions, pairs, pair_count, pitch, bucket_particle_count);
}

#endif

__global__ void add_interactions(particle_t* list, const int_t n_particles, int* interactions) {
    const int_t p_id = blockDim.x * blockIdx.x + threadIdx.x;
    if (p_id >= n_particles)
        return;
    atomicAdd((int*)&INTER(p_id), interactions[p_id]);
}

__global__ void reset_fn(particle_t* list, const int_t n_particles, real_t* subdomain) {
    const int_t k = blockDim.x * blockIdx.x + threadIdx.x;
    if (k >= n_particles)
        return;
    #ifdef BUCKET
        particle_t *particle = &list[k];
        int actual_x = MIN((int) (((particle->x[0] - subdomain[0])+RADIUS) / BUCKET_RADIUS) , dev_n_buckets_x-1);
        int actual_y = MIN((int) ((particle->x[1]+1.55f*H) / BUCKET_RADIUS),dev_n_buckets_y-1);
        particle->local_idx = k;
        particle->bucket_id = actual_y + dev_n_buckets_y * actual_x;
    #endif
    INTER(k) = WSUM(k) = AVRHO(k) = 0;
}

void
find_neighbors (int_t n_total, particle_t *dev_particles, pair_t * dev_pairs, real_t* dev_subdomain, unsigned int *dev_pair_count, int* dev_interactions) {
    double t_before, t_after;
    //t_before = MPI_Wtime();
    reset_fn<<< ceilDiv(n_total, THREADS_PER_BLOCK), THREADS_PER_BLOCK >>>(dev_particles, n_total, dev_subdomain);
    CUDA_CALL(cudaDeviceSynchronize());
    //t_after = MPI_Wtime();
    //t_setup += t_after - t_before;

    // Allocation and other types of t_setup

    // More setup for cell list methods
    #ifdef BUCKET
        t_before = MPI_Wtime();
        CUDA_CALL(cudaMemset(bucket_particle_count, 0, sizeof(unsigned int) * n_buckets));
        // Initialize buckets
        fill_buckets<<< ceilDiv(n_total, THREADS_PER_BLOCK), THREADS_PER_BLOCK>>>(n_total, dev_particles, buckets, pitch, bucket_particle_count, n_buckets);
        CUDA_CALL(cudaDeviceSynchronize());
        t_after = MPI_Wtime();
        t_setup += t_after - t_before;
    #endif

    // Compute
    t_before = MPI_Wtime();
    #ifdef BUCKET
        find_pairs<<< ceilDiv(n_total, THREADS_PER_BLOCK), THREADS_PER_BLOCK >>>(buckets, n_buckets, dev_particles, n_total, dev_interactions, dev_pairs, dev_pair_count, pitch, bucket_particle_count);
        CUDA_CALL(cudaDeviceSynchronize());
    #endif
    #if !defined(BUCKET)
        dim3 grid(ceilDiv(n_total, BLOCK_SIZE), ceilDiv(n_total, BLOCK_SIZE));
        dim3 block(BLOCK_SIZE, BLOCK_SIZE);

        naive<<< grid, block >>>(n_total, dev_particles, dev_pairs, dev_pair_count, dev_interactions);
        CUDA_CALL(cudaDeviceSynchronize());
    #endif
    t_after = MPI_Wtime();
    t_find_pairs += t_after - t_before;

    t_before = MPI_Wtime();
    // Copy back pair count
    CUDA_CALL(cudaMemcpy(&n_pairs, dev_pair_count, sizeof(unsigned int), cudaMemcpyDeviceToHost));

    t_after = MPI_Wtime();
    t_transfer += t_after - t_before;

    add_interactions<<< ceilDiv(n_total, THREADS_PER_BLOCK), THREADS_PER_BLOCK >>>(dev_particles, n_total, dev_interactions);
    CUDA_CALL(cudaDeviceSynchronize());
}

__global__ void ext_force (particle_t* list, int_t n_field) {
    const int_t p_id = blockDim.x * blockIdx.x + threadIdx.x;
    if (p_id >= n_field)
        return;
    EXDVXDT(p_id, 1) = -9.81f; // Gravity
}

__global__ void reset_indvxdt(particle_t* list, const int_t n_particles) {
    const int_t p_id = blockDim.x * blockIdx.x + threadIdx.x;
    if (p_id >= n_particles)
        return;
    INDVXDT(p_id, 0) = 0.0f;
    INDVXDT(p_id, 1) = 0.0f;
}

__global__ void set_density(particle_t* list, int_t n_field) {
    const int_t p_id = blockDim.x * blockIdx.x + threadIdx.x;
    if (p_id >= n_field)
        return;
    if (INTER(p_id) < free_surface)
            RHO(p_id) = density;
}

__global__ void state_equation(particle_t* list, const int_t n_particles) {
    const int_t k = blockDim.x * blockIdx.x + threadIdx.x;
    if (k >= n_particles)
        return;
    P(k) = sos * sos * density * ((POW7(RHO(k) / density) - 1.0f) / 7.0f);
}


__global__ void
__launch_bounds__(THREADS_PER_BLOCK)
pairwise_interaction (particle_t * list, pair_t * pairs, int_t n_pairs) {
    const int_t kk = blockDim.x * blockIdx.x + threadIdx.x;
    if (kk >= n_pairs)
        return;
    pair_t pair = pairs[kk];
    int_t
            i = pair.i,
            j = pair.j;
    real_t hx_i, hy_i, hx_j, hy_j;

    real_t dwdx0 = pair.dwdx[0];
    real_t dwdx1 = pair.dwdx[1];
    particle_t p_i = list[i];
    particle_t p_j = list[j];

    // i acts on j
    hx_i = (-(p_i.p / POW2(p_i.rho) + p_j.p / POW2(p_j.rho)) * dwdx0) * p_j.mass;
    hy_i = (-(p_i.p / POW2(p_i.rho) + p_j.p / POW2(p_j.rho)) * dwdx1) * p_j.mass;

    // j acts on i, reverse sign because dwdx is X(i)-X(j)
    hx_j = (-(p_j.p / POW2(p_j.rho) + p_i.p / POW2(p_i.rho)) * (-dwdx0)) * p_i.mass;
    hy_j = (-(p_j.p / POW2(p_j.rho) + p_i.p / POW2(p_i.rho)) * (-dwdx1)) * p_i.mass;

    atomicAdd(&INDVXDT(i, 0), hx_i);
    atomicAdd(&INDVXDT(i, 1), hy_i);
    atomicAdd(&INDVXDT(j, 0), hx_j);
    atomicAdd(&INDVXDT(j, 1), hy_j);
}

__global__ void
__launch_bounds__(THREADS_PER_BLOCK)
dev_correction (particle_t * list, pair_t * pairs, int_t n_pairs) {
    const int_t p_id = blockDim.x * blockIdx.x + threadIdx.x;
    if (p_id >= n_pairs)
        return;
    int_t
            i = pairs[p_id].i,
            j = pairs[p_id].j;
    real_t w_p = pairs[p_id].w;
    real_t drho_i, drho_j;
    real_t rho_i = RHO(i);
    real_t rho_j = RHO(j);
    real_t w_i = WSUM(i);
    real_t w_j = WSUM(j);

    drho_i = -(rho_i - rho_j) * w_p / w_i;
    drho_j = -(rho_j - rho_i) * w_p / w_j;

    atomicAdd(&AVRHO(i), drho_i);
    atomicAdd(&AVRHO(j), drho_j);
}

__global__ void correct_rho(particle_t* list, const int_t n_particles) {
    const int_t k = blockDim.x * blockIdx.x + threadIdx.x;
    if (k >= n_particles)
        return;
    if (TYPE(k) < 0 && INTER(k) < 10)
        RHO(k) = density;
    else
        RHO(k) += 0.5 * AVRHO(k);
}

__global__ void reset_density (particle_t * list, const int_t n_particles) {
    const int_t p_id = blockDim.x * blockIdx.x + threadIdx.x;
    if (p_id >= n_particles)
        return;
    DRHODT(p_id) = 0.0f;
}

__global__ void
__launch_bounds__(THREADS_PER_BLOCK)
dev_cont_density (particle_t * list, pair_t * pairs, int_t n_pairs) {
    const int_t p_id = blockDim.x * blockIdx.x + threadIdx.x;
    if (p_id >= n_pairs)
        return;
    int_t
        i = pairs[p_id].i,
        j = pairs[p_id].j;
    real_t vcc_i, vcc_j;

    real_t dwdx0 = pairs[p_id].dwdx[0];
    real_t dwdx1 = pairs[p_id].dwdx[1];
    real_t rho_i = RHO(i);
    real_t rho_j = RHO(j);
    real_t m_i = M(i);
    real_t m_j = M(j);
    real_t vx_i = VX(i);
    real_t vy_i = VY(i);
    real_t vx_j = VX(j);
    real_t vy_j = VY(j);

    vcc_i = (vx_i-vx_j)*dwdx0 +
          (vy_i-vy_j)*dwdx1;


    // Reverse sign of dwdx because it is calc. with X(i)-X(j)
    vcc_j = (vx_j-vx_i)*(-dwdx0) +
          (vy_j-vy_i)*(-dwdx1);

    atomicAdd(&DRHODT(i), rho_i * (m_j/rho_j) * vcc_i);
    atomicAdd(&DRHODT(j), rho_j * (m_i/rho_i) * vcc_j);
}

__global__ void add_rho(particle_t * list, const int_t n_particles) {
    const int_t p_id = blockDim.x * blockIdx.x + threadIdx.x;
    if (p_id >= n_particles)
        return;
    RHO(p_id) += 0.5f * dt * DRHODT(p_id);
}

__global__ void
__launch_bounds__(THREADS_PER_BLOCK)
kernel (particle_t * list, pair_t * pairs, const int_t n_pairs) {
    const int_t p_id = blockDim.x * blockIdx.x + threadIdx.x;
    if (p_id >= n_pairs)
        return;
    static real_t factor = 7.0f / (478.0f * M_PI * H * H);
    pair_t *p = &pairs[p_id];  // convenience alias
    real_t dx[2] = {X(p->i)-X(p->j), Y(p->i)-Y(p->j)};
    real_t dwdx[2], w;

    real_t q = p->q;
    real_t r = p->r;

    // Branchless multiple if statement
    w = factor * (
        (q >= 0.0f & q <= 1.0f)   * (POW5((3-q)) - 6*POW5((2-q)) + 15*POW5((1-q))) + 
        (q > 1.0f & q <= 2.0f)   * (POW5(3-q) - 6*POW5(2-q)) +
        (q > 2.0f & q <= 3.0f)   * (POW5(3-q))
    );

    dwdx[0] =
        (q > 0.0f & q <= 1.0f) * (factor/POW2(H)) * (-120+120*POW2(q)-50*POW3(q))*dx[0] + 
        (q > 1.0f & q <= 2.0f) * (factor/H) * ((-5)*POW4((3-q))+30*POW4((2-q)))*(dx[0]/r) +
        (q > 2.0f & q <= 3.0f) * (factor/H) * ((-5)*POW4((3-q)))*(dx[0]/r)
    ;

    dwdx[1] = 
        (q > 0.0f & q <= 1.0f) * (factor/POW2(H)) * (-120+120*POW2(q)-50*POW3(q))*dx[1] + 
        (q > 1.0f & q <= 2.0f) * (factor/H) * ((-5)*POW4((3-q))+30*POW4((2-q)))*(dx[1]/r) +
        (q > 2.0f & q <= 3.0f) * (factor/H) * ((-5)*POW4((3-q)))*(dx[1]/r)
    ;
    

    p->w = w;
    memcpy(p->dwdx, dwdx, sizeof(real_t)*2);
    atomicAdd(&WSUM(p->i), w);
    atomicAdd(&WSUM(p->j), w);
}

__global__ void field_movement(particle_t* list, int_t n_field) {
    const int_t p_id = blockDim.x * blockIdx.x + threadIdx.x;
    if (p_id >= n_field)
        return;
    DVX(p_id,0) = INDVXDT(p_id,0) + EXDVXDT(p_id,0); //+ ardvxdt
    DVX(p_id,1) = INDVXDT(p_id,1) + EXDVXDT(p_id,1); //+ ardvxdt
}

// Two versions of the same kernel, to avoid branching
__global__ void inital_distributions(particle_t* list, const int_t n_field) {
    const int_t k = blockDim.x * blockIdx.x + threadIdx.x;
    if (k >= n_field)
        return;
    real_t rho, drhodt, v[2], dvx[2], x[2];
    rho = RHO(k);
    drhodt = DRHODT(k);
    dvx[0] = DVX(k,0);
    dvx[1] = DVX(k,1);
    v[0] = VX(k);
    v[1] = VY(k);
    x[0] = X(k);
    x[1] = Y(k);
    __syncthreads();

    rho += 0.5f * dt * drhodt;
    v[0] += 0.5f * dt * dvx[0];
    v[1] += 0.5f * dt * dvx[1];
    x[0] += dt * v[0];
    x[1] += dt * v[1];

    __syncthreads();
    RHO(k) = rho;
    VX(k) = v[0];
    VY(k) = v[1];
    X(k) = x[0];
    Y(k) = x[1];
}

__global__ void distributions(particle_t* list, const int_t n_field) {
    const int_t k = blockDim.x * blockIdx.x + threadIdx.x;
    if (k >= n_field)
        return;
    real_t rho, drhodt, v[2], dvx[2], x[2];
    rho = RHO(k);
    drhodt = DRHODT(k);
    dvx[0] = DVX(k,0);
    dvx[1] = DVX(k,1);
    v[0] = VX(k);
    v[1] = VY(k);
    x[0] = X(k);
    x[1] = Y(k);
    __syncthreads();

    rho += 0.5f * dt * drhodt;
    v[0] += 0.5f * dt * dvx[0];
    v[1] += 0.5f * dt * dvx[1];

    // Reflect velocity at boundaries - branchless
    // if the particles goes too far down
    v[1] = v[1] * (
        ((x[1] < 0.0f & v[1] < 0.0f) * -2) + 1
    );
    // if the particle goes too far to the left or right
    v[0] = v[0] * (
        (((x[0] > B & v[0] > 0.0f) | (x[0] < 0.0f & v[0] < 0.0f)) * -2) + 1
    );

    x[0] += dt * v[0];
    x[1] += dt * v[1];

    __syncthreads();

    RHO(k) = rho;
    VX(k) = v[0];
    VY(k) = v[1];
    X(k) = x[0];
    Y(k) = x[1];
}

__global__ void step_field_particles(particle_t* list, int_t n_field) {
    const int_t k = blockDim.x * blockIdx.x + threadIdx.x;
    if (k >= n_field)
        return;
    real_t dvx[2], v[2], x[2];
    dvx[0] = DVX(k,0);
    dvx[1] = DVX(k,1);
    v[0] = VX(k);
    v[1] = VY(k);
    x[0] = X(k);
    x[1] = Y(k);

    __syncthreads();
    // Vx, Vy cloned to vx_min vy_min for some reason
    // Omitting this until I see the purpose
    v[0] += 0.5f * dt * dvx[0];
    v[1] += 0.5f * dt * dvx[1];
    x[0] += dt * v[0];
    x[1] += dt * v[1];

    __syncthreads();
    list[k].v[0] = v[0];
    list[k].v[1] = v[1];
    list[k].x[0] = x[0];
    list[k].x[1] = x[1];
}

extern "C" void time_step(int_t timestep, int_t n_field, int_t n_virt, int_t n_mirror, particle_t* list) {
    double t_before, t_after;
    // Initialize particles on device
    const int_t n_particles = n_field + n_virt + n_mirror;
    if (n_particles == 0)
        return;

    t_before = MPI_Wtime();
    if (n_particles > dev_particles_max) {
        dev_particles_max = n_particles;
        CUDA_CALL(cudaFree(dev_particles));
        CUDA_CALL(cudaFree(dev_pairs));
        CUDA_CALL(cudaFree(dev_interactions));
        CUDA_CALL(cudaMalloc(&dev_particles, dev_particles_max * sizeof(particle_t)));
        CUDA_CALL(cudaMalloc(&dev_pairs, PAIR_ESTIMATE * dev_particles_max * sizeof(pair_t)));
        CUDA_CALL(cudaMalloc(&dev_interactions, dev_particles_max * sizeof(int)));
    }
    CUDA_CALL(cudaMemset(dev_pairs, 0, PAIR_ESTIMATE * n_particles * sizeof(pair_t)));
    CUDA_CALL(cudaMemset(dev_pair_count, 0, sizeof(unsigned int)));
    t_after = MPI_Wtime();
    t_alloc += t_after - t_before;

    t_before = MPI_Wtime();
    CUDA_CALL(cudaMemcpy(dev_particles, list, n_particles * sizeof(particle_t), cudaMemcpyHostToDevice));
    t_after = MPI_Wtime();
    t_transfer += t_after - t_before;

    t_before = MPI_Wtime();
    CUDA_CALL(cudaMemset(dev_interactions, 0, n_particles *sizeof(int)));
    t_after = MPI_Wtime();
    t_alloc += t_after - t_before;

    if ( timestep > 0 ) {
        t_before = MPI_Wtime();
        step_field_particles<<< ceilDiv(n_field, THREADS_PER_BLOCK), THREADS_PER_BLOCK >>>(dev_particles, n_field);
        CUDA_CALL(cudaDeviceSynchronize());
        t_after = MPI_Wtime();
        t_physics += t_after - t_before;
    }
    
    t_before = MPI_Wtime();
    find_neighbors(n_particles, dev_particles, dev_pairs, dev_subdomain, dev_pair_count, dev_interactions);
    t_after = MPI_Wtime();
    t_find_neighbors += t_after - t_before;

    t_before = MPI_Wtime();

    ext_force<<< ceilDiv(n_field, THREADS_PER_BLOCK), THREADS_PER_BLOCK >>>(dev_particles, n_field);
    reset_indvxdt<<< ceilDiv(n_particles, THREADS_PER_BLOCK), THREADS_PER_BLOCK >>>(dev_particles, n_particles);

    if (n_pairs > 0)
        kernel<<< ceilDiv(n_pairs, THREADS_PER_BLOCK), THREADS_PER_BLOCK >>>(dev_particles, dev_pairs, n_pairs);

    // <Cont density>
    reset_density<<< ceilDiv(n_particles, THREADS_PER_BLOCK), THREADS_PER_BLOCK >>>(dev_particles, n_particles);
    CUDA_CALL(cudaDeviceSynchronize());

    if (n_pairs > 0){
        dev_cont_density<<< ceilDiv(n_pairs, THREADS_PER_BLOCK), THREADS_PER_BLOCK >>>(dev_particles, dev_pairs, n_pairs);
        CUDA_CALL(cudaDeviceSynchronize());
    }

    add_rho<<< ceilDiv(n_particles, THREADS_PER_BLOCK), THREADS_PER_BLOCK >>>(dev_particles, n_particles);
    CUDA_CALL(cudaDeviceSynchronize());
    // </Cont Density>

    // Correction
    if ( timestep > 0 ) {
        if (n_pairs > 0) {
            dev_correction<<< ceilDiv(n_pairs, THREADS_PER_BLOCK), THREADS_PER_BLOCK >>>(dev_particles, dev_pairs, n_pairs);
            CUDA_CALL(cudaDeviceSynchronize());
        }

        correct_rho<<< ceilDiv(n_particles, THREADS_PER_BLOCK), THREADS_PER_BLOCK>>>(dev_particles, n_particles);
        CUDA_CALL(cudaDeviceSynchronize());
    }

    // <int force>
    set_density<<< ceilDiv(n_field, THREADS_PER_BLOCK), THREADS_PER_BLOCK >>>(dev_particles, n_field);
    CUDA_CALL(cudaDeviceSynchronize());

    // Equations of state
    state_equation<<< ceilDiv(n_particles, THREADS_PER_BLOCK), THREADS_PER_BLOCK >>>(dev_particles, n_particles);
    CUDA_CALL(cudaDeviceSynchronize());

    if (n_pairs > 0) {
        // All the pairwise interactions
        pairwise_interaction<<< ceilDiv(n_pairs, THREADS_PER_BLOCK), THREADS_PER_BLOCK >>>(dev_particles, dev_pairs, n_pairs);
        CUDA_CALL(cudaDeviceSynchronize());
    }
    // </int force>
    

    field_movement<<< ceilDiv(n_field, THREADS_PER_BLOCK), THREADS_PER_BLOCK >>>(dev_particles, n_field);
    CUDA_CALL(cudaDeviceSynchronize());
    
    if (timestep == 0) {
        inital_distributions<<< ceilDiv(n_field, THREADS_PER_BLOCK), THREADS_PER_BLOCK >>>(dev_particles, n_field);
        CUDA_CALL(cudaDeviceSynchronize());
    } else {
        distributions<<< ceilDiv(n_field, THREADS_PER_BLOCK), THREADS_PER_BLOCK >>>(dev_particles, n_field);
        CUDA_CALL(cudaDeviceSynchronize());
    }

    t_after = MPI_Wtime();
    t_physics += t_after - t_before;

    // Copy back particles
    t_before = MPI_Wtime();
    CUDA_CALL(cudaMemcpy(list, dev_particles, sizeof(particle_t)*n_particles, cudaMemcpyDeviceToHost));
    t_after = MPI_Wtime();
    t_transfer += t_after - t_before;
}

extern "C" void initialize_cuda(int_t rank, int_t cell_total, real_t *subdomain) {
    int num_devices;
    CUDA_CALL(cudaGetDeviceCount(&num_devices));
    CUDA_CALL(cudaSetDevice(rank % num_devices));

    CUDA_CALL(cudaMalloc(&dev_pair_count, sizeof(unsigned int)));
    CUDA_CALL(cudaMalloc(&dev_subdomain, sizeof(real_t)*2));
    CUDA_CALL(cudaMalloc(&dev_interactions, dev_particles_max * sizeof(int)));
    CUDA_CALL(cudaMalloc(&dev_particles, dev_particles_max * sizeof(particle_t)));
    CUDA_CALL(cudaMalloc(&dev_pairs, PAIR_ESTIMATE * dev_particles_max * sizeof(pair_t)));


    CUDA_CALL(cudaMemcpy(dev_subdomain, subdomain, 2 * sizeof(real_t), cudaMemcpyHostToDevice));

    #ifdef BUCKET
        CUDA_CALL(cudaMemcpyToSymbol(dev_n_buckets_x, &n_buckets_x, sizeof(int_t)));
        CUDA_CALL(cudaMemcpyToSymbol(dev_n_buckets_y, &n_buckets_y, sizeof(int_t)));

        n_buckets = cell_total;
        // Create buckets on GPU
        // THIS CANNOT BE CALLED WITH ERROR CHECKING MACRO FOR SOME REASON
        cudaError_t __err = cudaMallocPitch(&buckets, &pitch, INITIAL_MAX_BUCKET*sizeof(unsigned int), n_buckets);
        if (__err != cudaSuccess) {
            printf("Cannot allocate cells, allocation error\n");
            exit(-1);
        }
        CUDA_CALL(cudaMalloc(&bucket_particle_count, sizeof(unsigned int) * n_buckets));
        CUDA_CALL(cudaMemset(bucket_particle_count, 0, sizeof(unsigned int) * n_buckets));
    #endif
}

extern "C" void finalize_cuda() {
    CUDA_CALL(cudaFree(dev_particles));
    CUDA_CALL(cudaFree(dev_pairs));
    CUDA_CALL(cudaFree(dev_pair_count));
    CUDA_CALL(cudaFree(dev_interactions));
    CUDA_CALL(cudaFree(dev_subdomain));
    #ifdef BUCKET
        CUDA_CALL(cudaFree(buckets));
        CUDA_CALL(cudaFree(bucket_particle_count));
    #endif
}

extern "C" void print_cuda_info(int_t rank) {
    int device;
    cudaGetDevice(&device);

    print_timing("Find neighbors\t%.4lf\t", "%.4lf\t", t_find_neighbors);
    print_timing("Cell setup\t%.4lf\t", "%.4lf\t", t_setup);
    print_timing("Physics step\t%.4lf\t", "%.4lf\t", t_physics);
    print_timing("Update pairs\t%.4lf\t", "%.4lf\t", t_update_pairs);
    print_timing("Mem transfer\t%.4lf\t", "%.4lf\t", t_transfer);
    print_timing("Allocation\t%.4lf\t", "%.4lf\t", t_alloc);
    print_timing("Find pairs\t%.4lf\t", "%.4lf\t", t_find_pairs);

    cudaDeviceProp devProp;
    CUDA_CALL(cudaGetDeviceProperties(&devProp, device));
    //printf("Rank %ld: %s, pciBusID: %d, pciDeviceId: %d\n", rank, devProp.name, devProp.pciBusID, devProp.pciDeviceID);
}
