#include "sph.h"

bool verbose = false, restart = false;
int size, rank, west, east;
real_t subdomain[2];
int_t n_field = 0,
        n_virt = 0,
        n_mirror = 0,
        n_global_field = 0;

/* Timing */
double
    t_generate = 0.0,
    t_border = 0.0,
    t_timestep = 0.0,
    t_migrate = 0.0,
    t_io = 0.0;

/* Run parameters */
int_t
    min_iteration = MIN_ITERATION_DEFAULT,
    max_iteration = MAX_ITERATION_DEFAULT,
    checkpoint_frequency = CHECKPOINT_FREQUENCY_DEFAULT;


particle_t *list;                  // Flat list of particle structs
int_t n_capacity = CAP_INCREMENT;   // Initial list capacity, grows

void print_timing(const char* full_string, const char* short_string, double value) {
    if (rank == 0) {
        printf(full_string, value);
        for (int r = 1; r < size; ++r) {
            size_t remote_string_size = strlen(short_string);
            char remote_string[remote_string_size];
            double remote_value = 0.0;
            MPI_Recv(remote_string, remote_string_size+1, MPI_CHAR, r, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
            MPI_Recv(&remote_value, 1, MPI_DOUBLE, r, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
            printf(remote_string, remote_value);
        }
        printf("\n");
    }
    else {
        size_t string_size = strlen(short_string);
        MPI_Send(short_string, string_size+1, MPI_CHAR, 0, 0, MPI_COMM_WORLD);
        MPI_Send(&value, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
    }
}


void
time_integration ( void )
{
    double t_start, t_end;
    /* Construct local list */
    int_t timestep;
    for ( timestep=min_iteration; timestep<max_iteration ; timestep++ )
    {
        #ifdef SKIP
        if (timestep % 2 == 0) {
        #endif
            // Reinitialize list of actuals
            n_field = n_particles();
            resize_list ( n_field );

            // Prepare list of local particles
            marshal_particles ( &(list[0]) );

            // Add ghosts, append to list
            // This calculates n_virt, so n_field+n_virt=n_total for now
            MPI_Barrier(MPI_COMM_WORLD);
            t_start = MPI_Wtime();
            generate_virtual_particles();
            t_end = MPI_Wtime();
            t_generate += t_end - t_start;

            // Synchronize with neighbors: mirror particles & mirror ghosts
            // Outcome is flat list, mirror particle count in n_mirror
            MPI_Barrier(MPI_COMM_WORLD);
            t_start = MPI_Wtime();
            border_exchange();
            t_end = MPI_Wtime();
            t_border += t_end - t_start;
        #ifdef SKIP
        }
        #endif

        // Node-local physics
        MPI_Barrier(MPI_COMM_WORLD);
        t_start = MPI_Wtime();
        time_step ( timestep, n_field, n_virt, n_mirror, list);
        t_end = MPI_Wtime();
        t_timestep += t_end - t_start;

        #ifdef SKIP
        if (timestep % 2 != 0) {
        #endif
            // Retrieve updates to actual particles from local list into hash tab
            unmarshal_particles( list, n_field );

            // Migrate particles moved across subdomain boundaries
            MPI_Barrier(MPI_COMM_WORLD);
            t_start = MPI_Wtime();
            migrate_particles();
            t_end = MPI_Wtime();
            t_migrate += t_end - t_start;
        #ifdef SKIP
        }
        #endif

        // Write field state to file every few iterations
#ifndef NO_IO
        if ((timestep % checkpoint_frequency) == 0) {
            MPI_Barrier(MPI_COMM_WORLD);
            t_start = MPI_Wtime();
            char filename[256];
            memset ( filename, 0, 256*sizeof(char) );
            sprintf ( filename, "plot/%.4"ISTR".dat",
                timestep/checkpoint_frequency
            );
            if ( rank == 0 )
                printf ( "Output at step %"ISTR", '%s'\n", timestep, filename );

            collect_checkpoint();
            write_checkpoint ( filename );
            t_end = MPI_Wtime();
            t_io += t_end - t_start;
        } else if (verbose && rank == 0)
            printf("Step %"ISTR"\n", timestep);
#endif //NO_IO

    }

    /* Print all timings */
    print_timing("Generate ghosts\t%.4lf\t", "%.4lf\t", t_generate);
    print_timing("Border exchange\t%.4lf\t", "%.4lf\t", t_border);
    print_timing("Time step\t%.4lf\t", "%.4lf\t", t_timestep);
    print_timing("Migration\t%.4lf\t", "%.4lf\t", t_migrate);
    print_timing("Input/output\t%.4lf\t", "%.4lf\t", t_io);

    // Print cuda info
    print_cuda_info(rank);

    /* Print shared variables */
    if (rank == 0) {
        printf ( "%ld iterations\n", timestep );
        printf ( "%ld particles\n", n_global_field );
        printf ( "%d ranks, OMP_NUM_THREADS=", size );
        int sys_retval = system ( "echo ${OMP_NUM_THREADS}" );
        printf ( "\t(returned %d)\n", sys_retval );
    }
}


int
main ( int argc, char **argv )
{
    MPI_Init ( &argc, &argv );
    MPI_Comm_rank ( MPI_COMM_WORLD, &rank );
    MPI_Comm_size ( MPI_COMM_WORLD, &size );
    options ( argc, argv );
    east = (rank + 1) % size;
    west = (rank + size - 1) % size;

    if ( !restart )
        initialize();
    else
        restart_checkpoint ( min_iteration );

    time_integration();
    finalize();

    MPI_Finalize();
}


void
generate_virtual_particles ( void )
{
    n_virt = 0;
    real_t boundary = 1.55f*H;

    // No particle adds more than 5 ghosts, make sure we have space
    resize_list(n_field * 5);

    // Loop over all actual particles, detect boundary interactions
    #pragma omp parallel for shared(n_virt)
    for ( int_t k=0; k<n_field; k++ )
    {
        int_t current_n_virt;

        // Horizontal mirror left
        if ( X(k) < boundary )
        {
            // Make use of one more list slot as ghost-k
            #pragma omp atomic capture
            current_n_virt = n_virt++;

            int_t gk = n_field + current_n_virt;
            X(gk) = -X(k), VX(gk) = -VX(k);
            Y(gk) = Y(k),  VY(gk) = VY(k);
            P(gk) = P(k), RHO(gk) = RHO(k), M(gk) = M(k);
            TYPE(gk) = -2, HSML(gk) = H;
        }

        // Horizontal mirror right
        if ( X(k) > B-boundary )
        {
            #pragma omp atomic capture
            current_n_virt = n_virt++;

            int_t gk = n_field + current_n_virt;
            X(gk) = 2*B-X(k), VX(gk) = -VX(k);
            Y(gk) = Y(k),     VY(gk) = VY(k);
            P(gk) = P(k), RHO(gk) = RHO(k);
            M(gk) = M(k); // Neumann boundary
            TYPE(gk) = -2, HSML(gk) = H;
        }

        // Vertical mirror bottom
        if ( Y(k) < boundary )
        {
            #pragma omp atomic capture
            current_n_virt = n_virt++;

            int_t gk = n_field + current_n_virt;
            X(gk) = X(k),  VX(gk) = VX(k);
            Y(gk) = -Y(k), VY(gk) = -VY(k);
            P(gk) = P(k), RHO(gk) = RHO(k), M(gk) = M(k);
            TYPE(gk) = -2, HSML(gk) = H;
        }

        // Lower left corner
        if ( X(k) < boundary && Y(k) < boundary )
        {
            #pragma omp atomic capture
            current_n_virt = n_virt++;

            int_t gk = n_field + current_n_virt;
            X(gk) = -X(k), VX(gk) = -VX(k);
            Y(gk) = -Y(k), VY(gk) = -VY(k);
            P(gk) = P(k), RHO(gk) = RHO(k), M(gk) = M(k);
            TYPE(gk) = -2, HSML(gk) = H;
        }

        // Lower right corner
        if ( X(k) > B-boundary && Y(k) < boundary )
        {
            #pragma omp atomic capture
            current_n_virt = n_virt++;

            int_t gk = n_field + current_n_virt;
            X(gk) = 2*B-X(k), VX(gk) = -VX(k);
            Y(gk) = -Y(k),    VY(gk) = -VY(k);
            P(gk) = P(k), RHO(gk) = RHO(k), M(gk) = M(k);
            TYPE(gk) = -2, HSML(gk) = H;
        }
    }
}


void
initialize ( void )
{

    /* Set up the hash table of actual (field) particles */
    particles_init ();

    /* Calculate the bounds of the local subdomain */
    real_t subdomain_size = B / (real_t)size;
    subdomain[0] = rank * subdomain_size;
    subdomain[1] = (rank+1) * subdomain_size;

    n_buckets_x = (int_t)(ceil(((subdomain[1]-subdomain[0])+2*RADIUS) / BUCKET_RADIUS));
    n_buckets_y = (int_t)(ceil(((1.5f*T)+1.55f*H) / BUCKET_RADIUS));

    if ( subdomain_size <= sqrt((scale_k*H)*(scale_k*H)) )
        fprintf ( stderr, "Rank %d: "
            "Warning, subdomain size smaller than interaction radius at "
            "scale %lf, results will not be correct.\n", rank, SCALE
        );

    /* Populate the local subdomain with any initial particles */
    int_t n[2] = { 1+L/DELTA, 1+T/DELTA };
    n_global_field = n[0]*n[1];

    for ( int_t i=0; i<n[1]; i++ )
    {
        for ( int_t j=0; j<n[0]; j++ )
        {
            int_t k = j * n[1] + i; // k is particle index
            real_t
                x = H + j * DELTA,
                y = H + i * DELTA;
            if ( x >= subdomain[0] && x < subdomain[1] )
            {
                particle_t *p = malloc(sizeof(particle_t));
                p->idx = k;
                p->x[0] = x;
                p->x[1] = y;
                p->v[0] = p->v[1] = 0.0f;
                p->rho = density;
                p->p = density * 9.81f * (T - p->x[1]);  // Hydrostatic pressure
                p->mass = L * T * density / (real_t)(n[0]*n[1]);
                p->type = 2;
                p->hsml = H;
                insert_particle ( p );
            }
        }
    }

    // Count local particles after initializeation
    n_field = n_particles();

    // Initial allocation for the lists of local particles and pairs
    list = (particle_t *) malloc ( n_capacity * sizeof(particle_t) );

    // Initialize cuda devices:
    initialize_cuda(rank, n_buckets_x*n_buckets_y, subdomain);
}


void
finalize ( void )
{
    particles_finalize ();
    free ( list );
    finalize_cuda();
}


/* MPI communication */


void
migrate_particles ( void )
{
    particle_t
        *actuals[n_field],
        *move_west[n_field], // Can't migrate more than whole set
        *move_east[n_field];
    int_t export_east = 0, export_west = 0, import_east = 0, import_west = 0;

    list_particles ( actuals );
    #pragma omp parallel for
    for ( int_t k=0; k<n_field; k++ )
    {
        particle_t *p = actuals[k];
        if ( p->x[0] < subdomain[0] && rank > 0 )
        {
            //remove_particle ( p );
            int k;
            #pragma omp atomic capture
            k = export_west++;
            move_west[k] = p;
        }
        else if ( p->x[0] > subdomain[1] && rank < (size-1) )
        {
            //remove_particle ( p );
            int k;
            #pragma omp atomic capture
            k = export_east++;
            move_east[k] = p;
        }
    }

    for ( int_t k=0; k<export_west; k++ )
        remove_particle(move_west[k]);
    for ( int_t k=0; k<export_east; k++ )
        remove_particle(move_east[k]);

    MPI_Sendrecv (
        &export_west, 1, INT_MACRO_MPI, west, 0,
        &import_east, 1, INT_MACRO_MPI, east, 0,
        MPI_COMM_WORLD, MPI_STATUS_IGNORE
    );
    MPI_Sendrecv (
        &export_east, 1, INT_MACRO_MPI, east, 0,
        &import_west, 1, INT_MACRO_MPI, west, 0,
        MPI_COMM_WORLD, MPI_STATUS_IGNORE
    );

    // Serialize in/out-bound particles
    particle_t *outlist = malloc (
        (export_west+export_east)*sizeof(particle_t)
    );
    particle_t *inlist = malloc (
        (import_west+import_east)*sizeof(particle_t)
    );

    // Outbound
    for ( int_t k=0; k<export_west; k++ )
        memcpy ( &(outlist[k]), move_west[k], sizeof(particle_t) );
    for ( int_t k=0; k<export_east; k++ )
        memcpy ( &(outlist[export_west+k]), move_east[k], sizeof(particle_t) );

    MPI_Sendrecv (
        &(outlist[0]),
        export_west*sizeof(particle_t), MPI_BYTE, west, 0,
        &(inlist[import_west]),
        import_east*sizeof(particle_t), MPI_BYTE, east, 0,
        MPI_COMM_WORLD, MPI_STATUS_IGNORE
    );
    MPI_Sendrecv (
        &(outlist[export_west]),
        export_east*sizeof(particle_t), MPI_BYTE, east, 0,
        &(inlist[0]),
        import_west*sizeof(particle_t), MPI_BYTE, west, 0,
        MPI_COMM_WORLD, MPI_STATUS_IGNORE
    );

    for ( int_t k=0; k<(import_west+import_east); k++ )
    {
        particle_t *p = malloc ( sizeof(particle_t) );
        memcpy ( p, &(inlist[k]), sizeof(particle_t) );
        insert_particle ( p );
    }

    free ( outlist );
    free ( inlist );

    // Remove migrated particles from local responsibility
    for ( int_t k=0; k<export_west; k++ )
        free ( move_west[k] );
    for ( int_t k=0; k<export_east; k++ )
        free ( move_east[k] );
}


void
border_exchange ( void )
{
    int_t
        export_east = 0, export_west = 0,
        import_east = 0, import_west = 0;

    // Scan the list and count particles within neighbor reach
    #pragma omp parallel for
    for ( int_t k=0; k<(n_field+n_virt); k++ )
    {
        if ( (X(k) - subdomain[0]) < scale_k*H && rank>0 )
            #pragma omp atomic update
            export_west += 1;
        if ( (subdomain[1] - X(k)) < scale_k*H && rank<size-1)
            #pragma omp atomic update
            export_east += 1;
    }

    MPI_Sendrecv (
        &export_west, 1, INT_MACRO_MPI, west, 0,
        &import_east, 1, INT_MACRO_MPI, east, 0,
        MPI_COMM_WORLD, MPI_STATUS_IGNORE
    );
    MPI_Sendrecv (
        &export_east, 1, INT_MACRO_MPI, east, 0,
        &import_west, 1, INT_MACRO_MPI, west, 0,
        MPI_COMM_WORLD, MPI_STATUS_IGNORE
    );

    // This transfer list could be glob/resize instead of malloc per iter

    particle_t *transfer = (particle_t *) malloc (
        (export_west + export_east) * sizeof(particle_t)
    );
    int_t w_idx = 0, e_idx=export_west;
    int_t priv_w_idx,  priv_e_idx;

    #pragma omp parallel for private(priv_w_idx, priv_e_idx)
    for ( int_t k=0; k<(n_field+n_virt); k++ )
    {
        if ( (X(k) - subdomain[0]) < scale_k*H && rank > 0)
        {
            #pragma omp atomic capture
            priv_w_idx = w_idx++;
            memcpy ( &(transfer[priv_w_idx]), &(list[k]), sizeof(particle_t) );
        }
        if ( (subdomain[1] - X(k)) < scale_k*H && rank < size-1)
        {
            #pragma omp atomic capture
            priv_e_idx = e_idx++;
            memcpy ( &(transfer[priv_e_idx]), &(list[k]), sizeof(particle_t) );
        }
    }

    n_mirror = import_east + import_west;
    resize_list ( n_field + n_virt + n_mirror );

    MPI_Sendrecv (
        &(transfer[0]),
        export_west*sizeof(particle_t), MPI_BYTE, west, 0,
        &(list[n_field+n_virt+import_west]),
        import_east*sizeof(particle_t), MPI_BYTE, east, 0,
        MPI_COMM_WORLD, MPI_STATUS_IGNORE
    );
    MPI_Sendrecv (
        &(transfer[export_west]),
        export_east*sizeof(particle_t), MPI_BYTE, east, 0,
        &(list[n_field+n_virt]),
        import_west*sizeof(particle_t), MPI_BYTE, west, 0,
        MPI_COMM_WORLD, MPI_STATUS_IGNORE
    );
    free ( transfer );
}

/* Auxiliary routines - file handling is in sph_io.c */

void
resize_list ( int_t required )
{
    if ( required > n_capacity )
    {
        n_capacity = required;
        particle_t* new_list = realloc ( list, n_capacity * sizeof(particle_t) );
        if(new_list == NULL) {
            free(list);
            fprintf(stderr, "1Not enough memory!\n");
            exit(1);
        } else {
            list = new_list;
        }
    }
}

void
options ( int argc, char **argv )
{
    /* Master rank parses command line options */
    if ( rank == 0 )
    {
        int o;
        while ( (o = getopt(argc,argv,"i:c:r:")) != -1 )
        switch ( o )
        {
            case 'i':
                max_iteration =
                    strtol(optarg, NULL, 10);
                break;
            case 'c':
                checkpoint_frequency =
                    strtol(optarg, NULL, 10);
                break;
            case 'r':
                min_iteration = strtol(optarg,NULL,10);
                break;
        }
    }

    /* Communicate option flags to the rest of the collective */
    MPI_Bcast ( &min_iteration, 1, INT_MACRO_MPI, 0,
        MPI_COMM_WORLD
    );
    MPI_Bcast ( &max_iteration, 1, INT_MACRO_MPI, 0,
        MPI_COMM_WORLD
    );
    MPI_Bcast ( &checkpoint_frequency, 1, INT_MACRO_MPI, 0,
        MPI_COMM_WORLD
    );
    if ( min_iteration != MIN_ITERATION_DEFAULT )
        restart = true;
}
